/*** DX11Device class, Cristiaan Kop							 updated 02-01-2017 ***/
/*** ============================================================================== ***/
/*** NAMESPACE: Crealysm::RENDERER (DX11)											***/
/*** cpp file for the DX11Device class implementation								***/

#include "RENDERER/DX11/DX11device.h"

#include <ceasylogging++.h>
#include "COMMON/crealysm_text.h"
#include "RENDERER/DX11/DX11helper.h"
#include "RENDERER/DX11/DX11rendertarget.h"
#include "RENDERER/DX11/DX11helper.h"

using namespace Crealysm::COMMON;


namespace Crealysm { namespace RENDERER
{

/**************************************************************************************/
/***								CONSTRUCTOR										***/
/*** ==> usage: when creating a DX11Device object									***/
/*** ==> create the device and context												***/
/**************************************************************************************/
DX11Device::DX11Device(HWND pHwnd, CRendererSettings *pSettings, const CR_BUFFER_FORMAT pBackBufferFormat, const CR_BUFFER_FORMAT pDepthBufferFormat) 
	: IDevice(pHwnd, pSettings, pBackBufferFormat, pDepthBufferFormat)
{
	mDXGIBackBufferFormat = GetDXGIFormat(mBackBufferFormat);
	mDXGIDepthBufferFormat = GetDXGIFormat(mDepthBufferFormat);
}

/**************************************************************************************/
/***								DESTRUCTOR										***/
/*** ==> usage: when DX11Device object is not needed anymore						***/
/*** ==> releases stuff, objects, memory etc. 										***/
/**************************************************************************************/

DX11Device::~DX11Device() 
{
	// nothing yet, using CComPtr
}

/**************************************************************************************/
/***								STARTUP 										***/
/*** ==> usage: creates the DX11 device, swapchain etc. 							***/
/*** ==> uses hardcoded driverType and BackBuffer format (not changed often)		***/
/**************************************************************************************/

bool DX11Device::Startup()
{
	// Create API device + context
	if(!Create()) return false;

	// Check if requested AA mode is supported
	if(mSettings->GetMSAAEnabled())
	{
		if(!CheckAASupport(mBackBufferFormat, mSettings->GetMSAANrSamples()))
		{
			CLOG(ERROR, "DX11RENDERER") << error_msaa_supported.c_str();
			CLOG(WARNING, "DX11RENDERER") << warning_msaa_disabled.c_str();
			mSettings->SetMSAAEnabled(false);
			mSettings->SetMSAANrSamples(0);
		}
	}

	// Create the SwapChain and DepthStencilBuffer
	if(!CreateSwapchain())
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_swapchain.c_str();	
		return false;
	}

	if(!CreateDepthBuffer())
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_ds.c_str();
		return false;
	}
	return true;
}

/**************************************************************************************/
/***										CREATE 									***/
/*** ==> usage: creates the DX11 device and device context 							***/
/*** ==> creates both API objects and checks feature level 		 					***/
/**************************************************************************************/

bool DX11Device::Create()
{
	D3D_DRIVER_TYPE driverType = D3D_DRIVER_TYPE_HARDWARE;
	uint createDeviceFlags = 0;

	#ifdef _DEBUG
    createDeviceFlags |= D3D11_CREATE_DEVICE_DEBUG;
	#endif

	D3D_FEATURE_LEVEL featureLevel;
	HRESULT hr = D3D11CreateDevice(0,                 // default adapter
								   driverType,		
								   0,                 // no software device
								   createDeviceFlags, 
								   0, 0,              // default feature level array
								   D3D11_SDK_VERSION,
								   &mDevice,
								   &featureLevel,
								   &mDeviceContext);

	if(FAILED(hr))
	{
		CLOG(FATAL, "DX11RENDERER") << fatal_create_device.c_str();
		return false;
	}

	if(featureLevel != D3D_FEATURE_LEVEL_11_0 )
	{
		CLOG(FATAL, "DX11RENDERER") << fatal_device_req.c_str();
		return false;
	}
	return true;
}

/**************************************************************************************/
/***							UPDATE FULLSCREEN STATE								***/
/*** ==> usage: changes the fullscreen mode 										***/
/*** ==> sets the SwapChain property fullscreen mode etc. 		 					***/
/**************************************************************************************/

bool DX11Device::UpdateFullscreenState()
{
	if(FAILED(mSwapChain->SetFullscreenState(!mSettings->GetWindowed(), nullptr))) return false;
	return true;
}

/**************************************************************************************/
/***									RESIZE										***/
/*** ==> usage: when the dimensions of the renderer are changed (1 viewport?)		***/
/*** ==> calls the private member functions for updating  		 					***/
/**************************************************************************************/

bool DX11Device::Resize()
{
	if(!ResizeSwapchain()) return false;
	if(!CreateDepthBuffer())
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_ds.c_str();
		return false;
	}
	return true;
}

/**************************************************************************************/
/***								CHECK AA SUPPORT								***/
/*** ==> usage: to check if the requested (MS)AA mode is supported on the GPU		***/
/*** ==> return if so and returns the by DX11 returned MSAA quality through param 	***/
/**************************************************************************************/

bool DX11Device::CheckAASupport(const CR_BUFFER_FORMAT pFormat, const uint pNrSamples)
{
	DXGI_FORMAT tempFormat = GetDXGIFormat(pFormat);
	uint quality;
	if(FAILED(mDevice->CheckMultisampleQualityLevels(tempFormat, pNrSamples, &quality))) return false;
	mSettings->SetAAQuality(quality);

	return true;
}

/**************************************************************************************/
/***								SET RENDERTARGET								***/
/*** ==> usage: to set a RT to the output merget state 								***/
/*** ==> calls the DX11 context to set the RT										***/
/**************************************************************************************/

bool DX11Device::SetRenderTarget(const IRenderTarget &pRT)
{
	ID3D11RenderTargetView *RTV[] = { (ID3D11RenderTargetView*)pRT.GetAPIViewPtr() };

	// Bind the render target view and depth/stencil view to the pipeline.
	mDeviceContext->OMSetRenderTargets(1, RTV, mDepthStencilView);
	return true;
}

/**************************************************************************************/
/***								RESET RENDERTARGETS								***/
/*** ==> usage: to set all RTs to 0 in the DX11 API								***/
/*** ==> resets the RTs that are currently bound to the stage	 					***/
/**************************************************************************************/

void DX11Device::ResetRenderTargets()
{
	mDeviceContext->OMSetRenderTargets(0, 0, 0);
}

/**************************************************************************************/
/***									PRESENT										***/
/*** ==> usage: to present the scene to a rendertarget 								***/
/*** ==> resolves RT and copyes results to BackBuffer			 					***/
/**************************************************************************************/

bool DX11Device::Present(const IRenderTarget &pRT)
{
	CComPtr<ID3D11Texture2D> backBuffer;
	if(FAILED(mSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), reinterpret_cast<void**>(&backBuffer)))) return false;
	mDeviceContext->ResolveSubresource(backBuffer, 0, (ID3D11Texture2D*)pRT.GetAPIBufferPtr(), 0, mDXGIBackBufferFormat);

	// Present the BackBuffer
	if(mSettings->GetVsync())
	{
		if(FAILED(mSwapChain->Present(1, 0))) return false;
	}
	else if(FAILED(mSwapChain->Present(0, 0))) return false;

	return true;
}

/**************************************************************************************/
/***								CLEAR RT VIEW	 								***/
/*** ==> usage: clears a RenderTarget view with a specific color					***/
/*** ==> n/a													  					***/
/**************************************************************************************/

bool DX11Device::ClearRTView(const IRenderTarget &pRT, const float pColor[4])
{
	mDeviceContext->ClearRenderTargetView((ID3D11RenderTargetView*)pRT.GetAPIViewPtr(), pColor);
	return true;
}

/**************************************************************************************/
/***								CLEAR DEPTH VIEW	 							***/
/*** ==> usage: clears the DepthStencilView											***/
/*** ==> n/a													  					***/
/**************************************************************************************/

bool DX11Device::ClearDepthView()
{
	mDeviceContext->ClearDepthStencilView(mDepthStencilView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 1.0f, 0);
	return true;
}

/**************************************************************************************/
/***								SET VIEWPORT	 								***/
/*** ==> usage: to set the current viewport on the device 							***/
/*** ==> sets the viewport(s) to the GPU device, DX11 API 		  					***/
/**************************************************************************************/

bool DX11Device::SetViewport(const CViewport &pViewport)
{
	D3D11_VIEWPORT tViewport;
	tViewport.TopLeftX	= (float)pViewport.GetTopLeftX();
	tViewport.TopLeftY	= (float)pViewport.GetTopLeftY();
	tViewport.Width		= static_cast<float>(pViewport.GetWidth());
	tViewport.Height	= static_cast<float>(pViewport.GetHeight());
	tViewport.MinDepth	= pViewport.GetMinDepth();
	tViewport.MaxDepth	= pViewport.GetMaxDepth();

	mDeviceContext->RSSetViewports(1, &tViewport);
	return true;
}

/**************************************************************************************/
/***								CREATE SWAPCHAIN 								***/
/*** ==> usage: creates the SwapChain												***/
/*** ==> creates the DX11 SwapChain object using swapchain desc  					***/
/**************************************************************************************/

bool DX11Device::CreateSwapchain()
{
	if(mSwapChain) return false;
	
	// Fill out the DXGI_SWAP_CHAIN_DESC
	DXGI_SWAP_CHAIN_DESC sd;
	sd.BufferDesc.Width						= mSettings->GetScreenWidth();
	sd.BufferDesc.Height					= mSettings->GetScreenHeight();
	sd.BufferDesc.RefreshRate.Numerator		= 60;
	sd.BufferDesc.RefreshRate.Denominator	= 1;
	sd.BufferDesc.Format					= mDXGIBackBufferFormat;
	sd.BufferDesc.ScanlineOrdering			= DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
	sd.BufferDesc.Scaling					= DXGI_MODE_SCALING_UNSPECIFIED;
	sd.SampleDesc.Count						= 1;									// MSAA through RenderTarget
	sd.SampleDesc.Quality					= 0;
	sd.BufferUsage							= DXGI_USAGE_RENDER_TARGET_OUTPUT;
	sd.BufferCount							= 1;
	sd.OutputWindow							= mHwnd;
	
	if(mSettings->GetWindowed()) sd.Windowed = true;
		else sd.Windowed = false;

	sd.SwapEffect   = DXGI_SWAP_EFFECT_DISCARD;
	sd.Flags        = DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH;

	// create swap chain using same IDXGIFactory as used to create device
	IDXGIDevice *dxgiDevice = nullptr;
	if(FAILED(mDevice->QueryInterface(__uuidof(IDXGIDevice), (void**)&dxgiDevice))) return false;

	IDXGIAdapter *dxgiAdapter = nullptr;
	if(FAILED(dxgiDevice->GetParent(__uuidof(IDXGIAdapter), (void**)&dxgiAdapter))) return false;

	IDXGIFactory *dxgiFactory = nullptr;
	if(FAILED(dxgiAdapter->GetParent(__uuidof(IDXGIFactory), (void**)&dxgiFactory))) return false;

	if(mSwapChain) mSwapChain = nullptr;
	if(FAILED(dxgiFactory->CreateSwapChain(mDevice, &sd, &mSwapChain))) return false;

	// DISABLE ALT+ENTER, fullscreen switching handled manually
	if(FAILED(dxgiFactory->MakeWindowAssociation(mHwnd, DXGI_MWA_NO_WINDOW_CHANGES))) return false;

	ReleaseCOM(dxgiDevice);
	ReleaseCOM(dxgiAdapter);
	ReleaseCOM(dxgiFactory);

	return true;
}

/**************************************************************************************/
/***								RESIZE SWAPCHAIN 								***/
/*** ==> usage: resizes the SwapChain												***/
/*** ==> calls the DXGI function to do so						 					***/
/**************************************************************************************/

bool DX11Device::ResizeSwapchain()
{
	if(FAILED(mSwapChain->ResizeBuffers(1, mSettings->GetScreenWidth(), mSettings->GetScreenHeight(), mDXGIBackBufferFormat, DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH))) return false;
	CLOG(DEBUG, "DX11RENDERER") << debug_resize_buffers.c_str() << mSettings->GetScreenWidth() << ", " << mSettings->GetScreenHeight() << ".";

	/*** NOTE: WHEN SWITCHING WINDOWED TO FULLSCREEN, WIDTH/HEIGHT IS AUTOMATICALLY ADJUSTED TO CLOSEST SUPPORTED RESOLUTION ***/
	/*** IE: 1290x740 dragged -> WILL BE 1280x720 ***/

	return true;
}

/**************************************************************************************/
/***							CREATE DEPTHBUFFER	 								***/
/*** ==> usage: (re)creates the Depth(Stencil) buffer for DX11						***/
/*** ==> fills in the buffer and creates the DepthStencilView	 					***/
/**************************************************************************************/

bool DX11Device::CreateDepthBuffer()
{
	if(mDepthStencilView) mDepthStencilView = nullptr;

	D3D11_TEXTURE2D_DESC depthStencilDesc;
	
	depthStencilDesc.Width				= mSettings->GetScreenWidth();
	depthStencilDesc.Height				= mSettings->GetScreenHeight();
	depthStencilDesc.MipLevels			= 1;
	depthStencilDesc.ArraySize			= 1;
	depthStencilDesc.Format				= mDXGIDepthBufferFormat;
	depthStencilDesc.SampleDesc.Count   = 1;
	depthStencilDesc.SampleDesc.Quality = 0;
  
	if(mSettings->GetMSAAEnabled())
	{
		depthStencilDesc.SampleDesc.Count = mSettings->GetMSAANrSamples();

		if(mSettings->GetAAQuality() == 0) depthStencilDesc.SampleDesc.Quality = D3D11_STANDARD_MULTISAMPLE_PATTERN;
		else depthStencilDesc.SampleDesc.Quality = mSettings->GetAAQuality()-1;
	}

	depthStencilDesc.Usage          = D3D11_USAGE_DEFAULT;
	depthStencilDesc.BindFlags      = D3D11_BIND_DEPTH_STENCIL;
	depthStencilDesc.CPUAccessFlags = 0; 
	depthStencilDesc.MiscFlags      = 0;

	CComPtr<ID3D11Texture2D> depthStencilBuffer;
	if(FAILED(mDevice->CreateTexture2D(&depthStencilDesc, 0, &depthStencilBuffer))) return false;
	if(FAILED(mDevice->CreateDepthStencilView(depthStencilBuffer, 0, &mDepthStencilView))) return false;

	return true;
}




/**************************************************************************************/
/***							GET API DEVICE PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer to the DX11 GPU device		 				***/
/*** ==> returns a pointer to the DX11 device (D3D11)		 						***/
/**************************************************************************************/

void* DX11Device::GetAPIDevicePtr() const
{
	return mDevice;
}

/**************************************************************************************/
/***							GET API CONTEXT PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer to the DX11 device context	 				***/
/*** ==> returns a pointer to the DX11 device context		 						***/
/**************************************************************************************/

void* DX11Device::GetAPIContextPtr() const
{
	return mDeviceContext;
}



}
}
