/*** DX11 RenderTarget class, Cristiaan Kop						 updated 02-01-2017 ***/
/*** ============================================================================== ***/
/*** NAMESPACE: Crealysm::RENDERER (DX11)											***/
/*** cpp file for the DX11RenderTarget class implementation							***/

#include "RENDERER/DX11/DX11rendertarget.h"

#include <ceasylogging++.h>
#include "COMMON/crealysm_text.h"
#include "RENDERER/DX11/DX11helper.h"

using namespace Crealysm::COMMON;


namespace Crealysm { namespace RENDERER
{

/**************************************************************************************/
/***								CONSTRUCTOR  									***/
/*** ==> usage: creates a DX11 RenderTarget				 							***/
/*** ==> fully creates the RT, if it fails -> nullptr's								***/
/**************************************************************************************/

DX11RenderTarget::DX11RenderTarget(IDevice *pDevice, const uint pWidth, const uint pHeight, const CR_BUFFER_FORMAT pBufferFormat, const uint pNrAASamples, const uint pAAQuality, const bool pUseBaseConfig)
	: IRenderTarget(pDevice, pWidth, pHeight, pBufferFormat, pNrAASamples, pAAQuality, pUseBaseConfig)
{
	mDXGIFormat = GetDXGIFormat(mBufferFormat);

	if(mAASamples > 1) mAAEnabled = true;
	else mAAEnabled = false;

	Create();
}

/**************************************************************************************/
/***									DESTRUCTOR  								***/
/*** ==> usage: destroys the DX11 RenderTarget			 							***/
/*** ==> free's memory etc.															***/
/**************************************************************************************/

DX11RenderTarget::~DX11RenderTarget()
{
	// nothing to do
}

/**************************************************************************************/
/***									CREATE  									***/
/*** ==> usage: creates DX11 buffer/texture 2d and the Rendertarget view 			***/
/*** ==> creation using current properties 											***/
/**************************************************************************************/

bool DX11RenderTarget::Create()
{
	if(!CreateBuffer())
	{
		CLOG(ERROR, "DX11RENDERER") << error_rt_create_tex.c_str();
		mBuffer = nullptr;
		return false;
	}

	// Create view
	CComPtr<ID3D11Device> myDev = (ID3D11Device*)mDevicePtr->GetAPIDevicePtr();

	if(mView) mView = nullptr;
	if(FAILED(myDev->CreateRenderTargetView(mBuffer, 0, &mView)))
	{
		CLOG(FATAL, "DX11RENDERER") << error_rt_create_view.c_str();
		mView = nullptr;
		return false;
	}
	return true;
}

/**************************************************************************************/
/***								CREATE BUFFER  									***/
/*** ==> usage: creates the DX11 buffer/texture 2d for the rendertarget				***/
/*** ==> (re)creates the buffer														***/
/**************************************************************************************/

bool DX11RenderTarget::CreateBuffer()
{
	D3D11_TEXTURE2D_DESC renderTargetTexDesc;
	renderTargetTexDesc.Width				= mWidth;
	renderTargetTexDesc.Height				= mHeight;
	renderTargetTexDesc.MipLevels			= 1;
	renderTargetTexDesc.ArraySize			= 1;
	renderTargetTexDesc.Format				= mDXGIFormat;
	renderTargetTexDesc.Usage				= D3D11_USAGE_DEFAULT;
	renderTargetTexDesc.BindFlags			= D3D11_BIND_RENDER_TARGET; //| D3D11_BIND_SHADER_RESOURCE;
	renderTargetTexDesc.CPUAccessFlags		= 0; 
	renderTargetTexDesc.MiscFlags			= 0;
	renderTargetTexDesc.SampleDesc.Count	= 1;
	renderTargetTexDesc.SampleDesc.Quality	= 0;

	if(mAAEnabled)
	{
		renderTargetTexDesc.SampleDesc.Count	= mAASamples;
		renderTargetTexDesc.SampleDesc.Quality	= mAAQuality-1;
	}

	CComPtr<ID3D11Device> myDev = (ID3D11Device*)mDevicePtr->GetAPIDevicePtr();

	if(mBuffer) mBuffer = nullptr;
	if(FAILED(myDev->CreateTexture2D(&renderTargetTexDesc, 0, &mBuffer))) return false;
	return true;
}

/**************************************************************************************/
/***							RESIZE / CHANGE (AA)								***/
/*** ==> usage: changes the rendertarget's size and/or AA properties				***/
/*** ==> updates the properties, recreated the buffer and view 						***/
/**************************************************************************************/

bool DX11RenderTarget::Resize(const uint pWidth, const uint pHeight)
{
	if(!Change(pWidth, pHeight, mAASamples, mAAQuality)) return false;	
	return true;
}

bool DX11RenderTarget::ChangeAA(const uint pNrAASamples, const uint pAAQuality)
{
	if(!Change(mWidth, mHeight, pNrAASamples, pAAQuality)) return false;	
	return true;
}

bool DX11RenderTarget::Change(const uint pWidth, const uint pHeight, const uint pNrAASamples, const uint pAAQuality)
{
	mWidth		= pWidth;
	mHeight		= pHeight;
	mAASamples	= pNrAASamples;
	mAAQuality	= pAAQuality;

	if(mAASamples > 1) mAAEnabled = true;
	else mAAEnabled = false;

	if(!Create()) return false;
	return true;
}




/**************************************************************************************/
/***							GET API BUFFER PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer to the DX11 RTs buffer/texture 				***/
/*** ==> returns a pointer to the DX11 buffer object		 						***/
/**************************************************************************************/

void* DX11RenderTarget::GetAPIBufferPtr() const
{
	return mBuffer;
}

/**************************************************************************************/
/***							GET API VIEW PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer to the DX11 RTs RenderTargetView 			***/
/*** ==> returns a pointer to the DX11 RTV object									***/
/**************************************************************************************/

void* DX11RenderTarget::GetAPIViewPtr() const
{
	return mView;
}


}
}
