/*** DX11 shaderpack class, Cristiaan Kop						 updated 27-12-2016 ***/
/*** ============================================================================== ***/
/*** NAMESPACE: Crealysm::RENDERER (DX11)											***/
/*** cpp file with the DX11ShaderPack class implementation			     		    ***/

#include "RENDERER/DX11/DX11shaderpack.h"

#include <d3dcompiler.h>
#include <d3dcommon.h>
#include <md5.h>
#include <ciniparser.h>
#include <ceasylogging++.h>

#include "COMMON/Crealysm_text.h"

using namespace Crealysm::COMMON;


namespace Crealysm { namespace RENDERER
{

/**************************************************************************************/
/***								CONSTRUCTOR										***/
/*** ==> usage: creates the shaderpack, D3D11 shader objects, MD5 checksum etc		***/
/*** ==> sets up the ShaderPack for use												***/
/**************************************************************************************/

DX11ShaderPack::DX11ShaderPack() : mInputLayoutId(0), mIdentifier(0)
{
	mFlags = D3DCOMPILE_ENABLE_STRICTNESS;
	#if defined( DEBUG ) || defined( _DEBUG )
    mFlags |= D3DCOMPILE_DEBUG;
	#endif

	mElementDesc.resize(0);
}

/**************************************************************************************/
/***								DESTRUCTOR										***/
/*** ==> usage: when DX11ShaderPack object is not needed anymore					***/
/*** ==> releases stuff, COM objects, memory etc. 									***/
/**************************************************************************************/

DX11ShaderPack::~DX11ShaderPack()
{
	// nothing yet, all CComPtr's
}

/**************************************************************************************/
/***								CREATE SHADERS	 								***/
/*** ==> usage: create the shaders in the pack										***/
/*** ==> loads the data from shaderConfig, creates shaders and prepares inputlayout	***/
/**************************************************************************************/

bool DX11ShaderPack::CreateShaders(ID3D11Device *pDevice, const std::string &pShaderFolder, const std::string &pFilename, const uint pPackId, const std::string &pTarget)
{
	if(pDevice == nullptr) return false;

	if(!LoadMetadata(pFilename, pShaderFolder, pPackId)) return false;

	// Create Vertex Shader
	if(strcmp(mVSEntryPoint.c_str(), "no") != 0)
	{
		if(!CompileAndCreateVS(pDevice, "vs_" + pTarget)) return false;
		CLOG(INFO, "DX11RENDERER") << info_vs_created.c_str() << ": " << mVSFilename << "," << mVSEntryPoint;
	}
	
	// Create Pixel Shader
	if(strcmp(mPSEntryPoint.c_str(), "no") != 0)
	{
		if(!CompileAndCreatePS(pDevice, "ps_"+ pTarget)) return false;
		CLOG(INFO, "DX11RENDERER") << info_ps_created.c_str() << ": " << mPSFilename << "," << mPSEntryPoint;
	}

	// Create Geometry shader
	if(strcmp(mGSEntryPoint.c_str(), "no") != 0)
	{
		if(!CompileAndCreateGS(pDevice, "gs_" + pTarget)) return false;
		CLOG(INFO, "DX11RENDERER") << info_gs_created.c_str() << ": " << mGSFilename << "," << mGSEntryPoint;
	}

	// InputLayout: create element desc array and MD5 checksum
	if(!CreateElementDesc()) return false;
	if(!CreateInputLayoutMD5Checksum()) return false;

	/***** HOW TO LOAD COMPILED SHADER TO BLOB??																		TODO???	*****/
	/***** http://sunandblackcat.com/tipFullView.php?l=eng&topicid=26&topic=DirectX11-Shaders-Compilation				TODO???	*****/

	return true;
}

/**************************************************************************************/
/***								CREATE FALLBACK 								***/
/*** ==> usage: create a fallback shaderpack										***/
/*** ==> creates the fallback PS and VS, no INI/ shader config needed				***/
/**************************************************************************************/

bool DX11ShaderPack::CreateFallback(ID3D11Device *pDevice, const std::string &pTarget, const std::string &pRootFolder)
{
	if(pDevice == nullptr) return false;

	// no defines/ shader macro's
	mVSFilename = pRootFolder + "Crealysm_fallback.hlsl";
	mVSEntryPoint = "VS_fallback";
	mPSFilename = pRootFolder + "Crealysm_fallback.hlsl";
	mPSEntryPoint = "PS_fallback";
	mGSFilename = "no";
	mGSEntryPoint = "no";
	mIdentifierText = "FALLBACK";

	// Create Vertex Shader
	if(!CompileAndCreateVS(pDevice, "vs_" + pTarget)) return false;

	// Create Pixel Shader
	if(!CompileAndCreatePS(pDevice, "ps_"+ pTarget)) return false;

	// InputLayout: create element desc array and MD5 checksum
	if(!CreateElementDesc()) return false;
	if(!CreateInputLayoutMD5Checksum()) return false;
	
	CLOG(DEBUG, "DX11RENDERER") << info_fallback_sp_created.c_str();

	return true;
}

/**************************************************************************************/
/***								GET INPUT LAYOUT 								***/
/*** ==> usage: to create the inputlayout that the VS needs							***/
/*** ==> creates inputlayout in local scope and returns it as result				***/
/**************************************************************************************/

CComPtr<ID3D11InputLayout> DX11ShaderPack::GetInputLayout(ID3D11Device *pDevice)
{
	if(pDevice == nullptr) return nullptr;

	if(!mVSShaderBlob) return nullptr;
	if(mElementDesc.size() < 1) return nullptr;

	// copy Crealysm elementDesc array to D3D element desc array
	std::vector<D3D11_INPUT_ELEMENT_DESC> descs(0);

	D3D11_INPUT_ELEMENT_DESC newDesc;
	for(size_t it=0;it<mElementDesc.size();++it)
	{
		newDesc = mElementDesc[it].ElementDesc;
		newDesc.SemanticName = mElementDesc[it].SemanticName;

		descs.push_back(newDesc);
	}

	// Try to create Input Layout
	CComPtr<ID3D11InputLayout> inputLayout = nullptr;
	if(FAILED(pDevice->CreateInputLayout(&descs[0], descs.size(), mVSShaderBlob->GetBufferPointer(), mVSShaderBlob->GetBufferSize(), &inputLayout)))
	{
		CLOG(DEBUG, "DX11RENDERER") << debug_inplayout_error.c_str();
		return nullptr;
	}
	CLOG(DEBUG, "DX11RENDERER") << debug_inplayout_created.c_str();
	
	return inputLayout;
}

/**************************************************************************************/
/***								SET INPUTLAYOUT ID								***/
/*** ==> usage: to set the corresponding/ unique VS inputlayout ID					***/
/*** ==> updates the member uint storing the unique ID								***/
/**************************************************************************************/

void DX11ShaderPack::SetInputLayoutId(const uint pId)
{
	mInputLayoutId = pId;
}

/**************************************************************************************/
/***								LOAD METADATA									***/
/*** ==> usage: to load the shaderpacks meta data from file						***/
/*** ==> parses the INI file and retrieves all shaderpack metadata 					***/
/**************************************************************************************/

bool DX11ShaderPack::LoadMetadata(const std::string &pFilename, const std::string &pShaderFolder, const uint pPackId)
{
	// Start parsing the shaderpack INI
	Crealysm::COMMON::CIniParser packParser;
	if(!packParser.StartReading(pFilename)) return false;

	uint nrAttrib = 0;
	if(!packParser.GotoBlockByName("ShaderPack", pPackId, nrAttrib)) return false;

	// Load defines from INI file
	while(true)
	{
		std::string attrName = packParser.GetNextAttributeName();
		if(strcmp(attrName.c_str(), "defines") == 0) break;
	}
	std::string definesAll = packParser.GetString();

	// Are there any defines?
	if(strcmp(definesAll.c_str(), "no") != 0)
	{
		if(!CreateDefines(definesAll)) return false;
	}

	// Read rest of shaderpack data from INI
	for(size_t meta=1;meta<nrAttrib;++meta)
	{
		std::string tlabel = packParser.GetNextAttributeName();
		
		if(strcmp(tlabel.c_str(), "vs_filename") == 0)		mVSFilename = pShaderFolder + packParser.GetString();
		if(strcmp(tlabel.c_str(), "vs_entrypoint") == 0)	mVSEntryPoint = packParser.GetString();
		if(strcmp(tlabel.c_str(), "ps_filename") == 0)		mPSFilename = pShaderFolder + packParser.GetString();
		if(strcmp(tlabel.c_str(), "ps_entrypoint") == 0)	mPSEntryPoint = packParser.GetString();
		if(strcmp(tlabel.c_str(), "gs_filename") == 0)		mGSFilename = pShaderFolder + packParser.GetString();
		if(strcmp(tlabel.c_str(), "gs_entrypoint") == 0)	mGSEntryPoint = packParser.GetString();
		if(strcmp(tlabel.c_str(), "id_text") == 0)			mIdentifierText = packParser.GetString();
		if(strcmp(tlabel.c_str(), "id_numeric") == 0)		mIdentifier = packParser.GetInt();
	}
	return true;
}

/**************************************************************************************/
/***						CREATE INPUTLAYOUT MD5 CHECKSUM							***/
/*** ==> usage: to create the MD5 checksum of the VS inputlayout					***/
/*** ==> stores the created checksum in the class object							***/
/**************************************************************************************/

bool DX11ShaderPack::CreateInputLayoutMD5Checksum()
{
	if(!mVS) return false;
	if(mElementDesc.size() < 1) return false;

	unsigned char MD5CheckSum[16];
	MD5_Checksum((unsigned char*) &mElementDesc[0], sizeof(DX11_VS_INPUT_ELEMENTDESC) * mElementDesc.size(), MD5CheckSum);

	mInputLayoutMD5Checksum = std::string(reinterpret_cast<char*>(MD5CheckSum),16);
	
	CLOG(DEBUG, "DX11RENDERER") << debug_inplayout_md5.c_str() << " : " << mInputLayoutMD5Checksum;
	return true;
}

/**************************************************************************************/
/***								CREATE ELEMENT DESC								***/
/*** ==> usage: to create the array of inputlayout element descriptions				***/
/*** ==> create the Crealysm elementDesc objects, based on D3D11 elementDescs		***/
/**************************************************************************************/

bool DX11ShaderPack::CreateElementDesc()
{
	if(!mVSShaderBlob) return false;		// NOT nullptr?

	ID3D11ShaderReflection* vertexShaderReflection = nullptr;
	if(FAILED(D3DReflect(mVSShaderBlob->GetBufferPointer(), mVSShaderBlob->GetBufferSize(), IID_ID3D11ShaderReflection, (void**) &vertexShaderReflection)))
	{
		CLOG(DEBUG, "DX11RENDERER") << debug_d3dreflect_nok.c_str();
		return false;
	}

	// Get shader info
	D3D11_SHADER_DESC shaderDesc;
	vertexShaderReflection->GetDesc(&shaderDesc);

	// Read input layout description from shader info
	mElementDesc.resize(0);
	mElementDesc.reserve(shaderDesc.InputParameters);

	for(uint i=0;i<shaderDesc.InputParameters;++i)
	{
		D3D11_SIGNATURE_PARAMETER_DESC paramDesc;
		vertexShaderReflection->GetInputParameterDesc(i, &paramDesc);
 
		// fill out input element desc
		D3D11_INPUT_ELEMENT_DESC elementDesc;
		elementDesc.SemanticName = paramDesc.SemanticName;
		elementDesc.SemanticIndex = paramDesc.SemanticIndex;
		elementDesc.InputSlot = 0;
		elementDesc.AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
		elementDesc.InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elementDesc.InstanceDataStepRate = 0;
		elementDesc.Format = GetDXGIFormat(paramDesc.Mask, paramDesc.ComponentType);
 
		//save element desc
		DX11_VS_INPUT_ELEMENTDESC newDesc(elementDesc);
		mElementDesc.push_back(newDesc);
	}

	// Free shader reflection memory
	ReleaseCOM(vertexShaderReflection);
	return true;
}

/**************************************************************************************/
/***								CREATE DEFINES									***/
/*** ==> usage: to create the array of D3D macro defines for shader compiling		***/
/*** ==> extracts the names and values for the defines								***/
/**************************************************************************************/

bool DX11ShaderPack::CreateDefines(const std::string &pDefineLine)
{
	std::vector<uint> defineNameStartPos;
	std::vector<uint> defineNameEndPos;
	std::vector<uint> defineValStartPos;
	std::vector<uint> defineValEndPos;

	defineNameStartPos.push_back(0);		// first define starts at pos 0

	for(size_t i=0;i<pDefineLine.length();++i)
	{
		if(pDefineLine[i] == ':')
		{
			defineNameEndPos.push_back(i);
			defineValStartPos.push_back(i+1);
		}
		if(pDefineLine[i] == ';')
		{
			defineNameStartPos.push_back(i+1);
			defineValEndPos.push_back(i);
		}
	}
	
	// found any?
	if(defineNameStartPos.size() == 1) return false;
	
	// define line always ends with ';'
	defineNameStartPos.erase(defineNameStartPos.end() - 1);

	// push the names and values into the member std::vector
	mDefines.resize(0);

	C_SHADER_MACRO newDefine;
	for(size_t defs=0;defs<defineNameStartPos.size();++defs)
	{
		newDefine.Name = pDefineLine.substr(defineNameStartPos[defs], defineNameEndPos[defs] - defineNameStartPos[defs]).c_str();
		newDefine.Definition = pDefineLine.substr(defineValStartPos[defs], defineValEndPos[defs] - defineValStartPos[defs]).c_str();

		mDefines.push_back(newDefine);
	}
	return true;
}

/**************************************************************************************/
/***							COMPILE AND CREATE VS								***/
/*** ==> usage: to create and compile the VertexShader in the shaderpack			***/
/*** ==> compiles the HLSL code to a D3D blob and creates shader object				***/
/**************************************************************************************/

bool DX11ShaderPack::CompileAndCreateVS(ID3D11Device *pDevice, const std::string &pTarget)
{
	// COMPILE: VS blob is stored as member, for Inputlayout usage later
	mVSShaderBlob = nullptr;
	ID3DBlob *errorBlob = nullptr;

	// create the D3D shader macro defines
	std::vector<D3D_SHADER_MACRO> defines = GetD3dDefines();

	// compile the shaderblob
	std::wstring loadfile(mVSFilename.begin(), mVSFilename.end());
	if(FAILED(D3DCompileFromFile(loadfile.c_str(), defines.data(), D3D_COMPILE_STANDARD_FILE_INCLUDE, mVSEntryPoint.c_str(), pTarget.c_str(), mFlags, 0, &mVSShaderBlob, &errorBlob)))
	{
		if(errorBlob)
		{
			CLOG(DEBUG, "DX11RENDERER") << error_d3dcompile.c_str() << " : " << (char*)errorBlob->GetBufferPointer();
			ReleaseCOM(errorBlob);
		}
        if(mVSShaderBlob) mVSShaderBlob = nullptr;
		return false;
	}
	ReleaseCOM(errorBlob);

	// create the D3D11 vertexshader object
	mVS = nullptr;
	if(FAILED(pDevice->CreateVertexShader(mVSShaderBlob->GetBufferPointer(), mVSShaderBlob->GetBufferSize(), nullptr, &mVS))) 
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_VS.c_str() << ", " << mVSFilename;
		return false;
	}
	return true;
}

/**************************************************************************************/
/***							COMPILE AND CREATE PS								***/
/*** ==> usage: to create and compile the PixelShader in the shaderpack				***/
/*** ==> compiles the HLSL code to a D3D blob and creates shader object				***/
/**************************************************************************************/

bool DX11ShaderPack::CompileAndCreatePS(ID3D11Device *pDevice, const std::string &pTarget)
{
	ID3D10Blob *errorBlob = nullptr;
	ID3D10Blob *shaderBlob = nullptr;

	// create the D3D shader macro defines
	std::vector<D3D_SHADER_MACRO> defines = GetD3dDefines();

	// compile the shaderblob
	std::wstring loadfile(mPSFilename.begin(), mPSFilename.end());
	if(FAILED(D3DCompileFromFile(loadfile.c_str(), defines.data(), D3D_COMPILE_STANDARD_FILE_INCLUDE, mPSEntryPoint.c_str(), pTarget.c_str(), mFlags, 0, &shaderBlob, &errorBlob)))
	{
		if(errorBlob)
		{
			CLOG(DEBUG, "DX11RENDERER") << error_d3dcompile.c_str() << " : " << (char*)errorBlob->GetBufferPointer();
			errorBlob->Release();
		}
        if(shaderBlob) shaderBlob->Release();
		return false;
	}
	ReleaseCOM(errorBlob);

	// create the D3D11 pixelshader object
	mPS = nullptr;
	if(FAILED(pDevice->CreatePixelShader(shaderBlob->GetBufferPointer(), shaderBlob->GetBufferSize(), nullptr, &mPS)))
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_PS.c_str() << ", " << mPSFilename;
		return false;
	}
	ReleaseCOM(shaderBlob); 
	return true;
}

/**************************************************************************************/
/***							COMPILE AND CREATE GS								***/
/*** ==> usage: to create and compile the GeometryShader in the shaderpack			***/
/*** ==> compiles the HLSL code to a D3D blob and creates shader object				***/
/**************************************************************************************/

bool DX11ShaderPack::CompileAndCreateGS(ID3D11Device *pDevice, const std::string &pTarget)
{
	ID3D10Blob *errorBlob = nullptr;
	ID3D10Blob *shaderBlob = nullptr;

	// create the D3D shader macro defines
	std::vector<D3D_SHADER_MACRO> defines = GetD3dDefines();

	// compile the shaderblob
	std::wstring loadfile(mGSFilename.begin(), mGSFilename.end());
	if(FAILED(D3DCompileFromFile(loadfile.c_str(), defines.data(), D3D_COMPILE_STANDARD_FILE_INCLUDE, mGSEntryPoint.c_str(), pTarget.c_str(), mFlags, 0, &shaderBlob, &errorBlob)))
	{
		if(errorBlob)
		{
			CLOG(DEBUG, "DX11RENDERER") << error_d3dcompile.c_str() << " : " << (char*)errorBlob->GetBufferPointer();
			errorBlob->Release();
		}
        if(shaderBlob) shaderBlob->Release();
		return false;
	}
	ReleaseCOM(errorBlob);

	// create the D3D11 geometryshader object
	mGS = nullptr;
	if(FAILED(pDevice->CreateGeometryShader(shaderBlob->GetBufferPointer(), shaderBlob->GetBufferSize(), nullptr, &mGS)))
	{
		CLOG(DEBUG, "DX11RENDERER") << error_create_GS.c_str() << ", " << mGSFilename;
		return false;
	}
	ReleaseCOM(shaderBlob);
	return true;
}




/**************************************************************************************/
/***								GET IDENTIFIER							  CONST	***/
/*** ==> usage: to retrieve the unique unumeric ID of this shaderpack				***/
/*** ==> returns the unique ID			 											***/
/**************************************************************************************/

uint DX11ShaderPack::GetIdentifier() const
{
	return mIdentifier;
}

/**************************************************************************************/
/***								GET IDENTIFIER TEXT						  CONST	***/
/*** ==> usage: to retrieve the unique text ID of this shaderpack					***/
/*** ==> returns the unique text ID		 											***/
/**************************************************************************************/

const std::string& DX11ShaderPack::GetIdentifierText() const
{
	return mIdentifierText;
}

/**************************************************************************************/
/***									HAS VS								  CONST	***/
/*** ==> usage: to check if the shaderpack has a VertexShader						***/
/*** ==> returns true if so				 											***/
/**************************************************************************************/

bool DX11ShaderPack::HasVS() const
{
	return(mVS != nullptr);
}

/**************************************************************************************/
/***									HAS PS								  CONST	***/
/*** ==> usage: to check if the shaderpack has a PixelShader						***/
/*** ==> returns true if so				 											***/
/**************************************************************************************/

bool DX11ShaderPack::HasPS() const
{
	return(mPS != nullptr);
}

/**************************************************************************************/
/***									HAS GS								  CONST	***/
/*** ==> usage: to check if the shaderpack has a GeometryShader						***/
/*** ==> returns true if so				 											***/
/**************************************************************************************/

bool DX11ShaderPack::HasGS() const
{
	return(mGS != nullptr);
}

/**************************************************************************************/
/***							GET VS SHADER BLOB	 					  	  CONST	***/
/*** ==> usage: to retrieve a const pointer to the VS's shader blob 				***/
/*** ==> returns the pointer, i.e. for creating inputlayouts (Readonly)				***/
/**************************************************************************************/

const ID3DBlob* DX11ShaderPack::GetVSShaderBlob() const
{
	return mVSShaderBlob;
}

/**************************************************************************************/
/***									GET VS PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer (const) to the D3D11 vertexshader object		***/
/*** ==> returns the pointer: both content and what it points to is readonly		***/
/**************************************************************************************/

ID3D11VertexShader* DX11ShaderPack::GetVSPtr() const
{
	return mVS;
}

/**************************************************************************************/
/***									GET PS PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer (const) to the D3D11 pixelshader object		***/
/*** ==> returns the pointer: both content and what it points to is readonly		***/
/**************************************************************************************/

ID3D11PixelShader* DX11ShaderPack::GetPSPtr() const
{
	return mPS;
}

/**************************************************************************************/
/***									GET GS PTR							  CONST	***/
/*** ==> usage: to retrieve a pointer (const) to the D3D11 geometryshader object	***/
/*** ==> returns the pointer: both content and what it points to is readonly		***/
/**************************************************************************************/

ID3D11GeometryShader* DX11ShaderPack::GetGSPtr() const
{
	return mGS;
}

/**************************************************************************************/
/***								GET INPUTLAYOUT ID						  CONST	***/
/*** ==> usage: to retrieve the unique ID for the corresponding VS inputlayout		***/
/*** ==> returns the ID																***/
/**************************************************************************************/

uint DX11ShaderPack::GetInputLayoutID() const
{
	return mInputLayoutId;
}

/**************************************************************************************/
/***						GET INPUTLAYOUT MD5 CHECKSUM					  CONST ***/
/*** ==> usage: retrieve the MD5 hash checksum for the VSs inputlayout				***/
/*** ==> returns a const ref to the member string (with checksum)					***/
/**************************************************************************************/

const std::string &DX11ShaderPack::GetInputLayoutMD5Checksum() const
{
	return mInputLayoutMD5Checksum;
}

/**************************************************************************************/
/***								GET D3D DEFINES						 	  CONST	***/
/*** ==> usage: to retrieve a std::vector with the D3D11 formatted macro's			***/
/*** ==> fills a D3D_SHADER_MACRO vector and returns is								***/
/**************************************************************************************/

std::vector<D3D_SHADER_MACRO> DX11ShaderPack::GetD3dDefines() const
{
	// create the D3D shader macro defines
	std::vector<D3D_SHADER_MACRO> defines;
	defines.reserve(mDefines.size() + 1);

	for(size_t m=0;m<mDefines.size();++m)
	{
		defines.push_back(D3D_SHADER_MACRO { mDefines[m].Name.c_str(), mDefines[m].Definition.c_str() });
	}
	defines.push_back(D3D_SHADER_MACRO { nullptr, nullptr });

	return defines;
}


}
}